import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:flutter_slick/navigation/route_builder.dart';
import 'package:provider/provider.dart';

import 'route_path.dart';
import 'router_delegate.dart';
import 'route_information_parser.dart';

void navigate(BuildContext context, RoutePath path) {
  Provider.of<NavigationLogic>(context, listen: false).setRoutePath(path);
}

class NavigationLogic extends ChangeNotifier {
  final BuildContext context;
  final Widget Function() initialScreen;
  final Widget Function() notFoundScreen;
  final RouteBuilder rootPath;
  final Stream? authState;

  final List<RoutePath> _history = [];

  late StandardRouterDelegate routerDelegate;
  late StandardRouteInformationParser routerInformationParser;

  late Widget _currentScreen;
  late RoutePath _routePath;

  RoutePath? _rewrittenRoutePath;

  StreamSubscription? _onAuthStateChangedSubscription;

  RoutePath? get routePath => _routePath;

  bool _isLoading = true;
  bool _isInitialising = true;

  bool get isLoading => _isLoading;

  set isLoading(bool value) {
    _isLoading = value;
    notifyListeners();
  }

  Widget get currentScreen => _currentScreen;

  RouteBuilder bareBuilder(
    String path,
    Widget Function() screen,
  ) {
    return RouteBuilder(
      pattern: RegExp(''),
      path: (_, __) => path,
      parser: RouteArguments.parse,
      build: (_, __, ___) async => screen(),
      rewrite: (_, __) async => null,
    );
  }

  NavigationLogic({
    required this.context,
    required this.initialScreen,
    required this.notFoundScreen,
    required this.rootPath,
    this.authState,
  }) : _currentScreen = initialScreen() {
    _routePath = bareBuilder('', initialScreen)();
    routerDelegate = StandardRouterDelegate(context, this);
    routerInformationParser = StandardRouteInformationParser(this);
    final authState = this.authState;
    if (authState != null) {
      _onAuthStateChangedSubscription = authState.listen(_onAuthStateChanged);
    } else {
      _isInitialising = false;
    }
  }

  Future<bool> popRoute() async {
    if (_history.isEmpty) {
      return false;
    } else {
      await setRoutePath(_history.last);
      return true;
    }
  }

  Future<void> setRoutePath(RoutePath routePath) async {
    if (_isInitialising) {
      _routePath = routePath;
      _currentScreen = initialScreen();
    } else {
      isLoading = true;
      // Adjust route path history
      if (_history.isNotEmpty && _history.last == routePath) {
        _history.removeLast();
      } else {
        if (_routePath != routePath) {
          _history.add(_routePath);
        }
      }
      // Detemine if a redirect is indicated
      RoutePath? currentPath = routePath;
      _rewrittenRoutePath = null;
      while (currentPath != null) {
        var result = await currentPath.rewrite(context);
        if (result == null) {
          _routePath = currentPath;
        } else {
          _rewrittenRoutePath = result;
        }
        currentPath = result;
      }
      // ignore: use_build_context_synchronously
      _currentScreen = await _routePath.build(context);
      isLoading = false;
    }
  }

  Future<void> _onAuthStateChanged(_) async {
    _isInitialising = false;
    await setRoutePath(_rewrittenRoutePath ?? _routePath);
  }

  @override
  void dispose() {
    _onAuthStateChangedSubscription?.cancel();
    super.dispose();
  }
}
